@@ -25,6 +25,8 @@
 #include <stdio.h>
 
 #include "ctkbanner.h"
+#include "common-utils.h"
+#include "ctkutils.h"
 
 /* pixdata headers */
 
@@ -39,7 +41,6 @@
 #include "color_pixdata.h"
 #include "config_pixdata.h"
 #include "crt_pixdata.h"
-#include "cursor_shadow_pixdata.h"
 #include "dfp_pixdata.h"
 #include "display_config_pixdata.h"
 #include "framelock_pixdata.h"
@@ -51,11 +52,12 @@
 #include "gvi_pixdata.h"
 #include "sdi_pixdata.h"
 #include "sdi_shared_sync_bnc_pixdata.h"
+#include "server_licensing_pixdata.h"
 #include "slimm_pixdata.h"
 #include "solaris_pixdata.h"
 #include "thermal_pixdata.h"
-#include "tv_pixdata.h"
 #include "vcs_pixdata.h"
+#include "vdpau_pixdata.h"
 #include "x_pixdata.h"
 #include "xvideo_pixdata.h"
 #include "svp_3dvp_pixdata.h"
@@ -66,12 +68,22 @@ ctk_banner_class_init    (CtkBannerClass *);
 
 static void
 ctk_banner_finalize      (GObject *);
+#ifdef CTK_GTK3
+static gboolean
+ctk_banner_draw_event  (GtkWidget *, cairo_t *);
 
+static void
+ctk_banner_get_preferred_width (GtkWidget *, gint *, gint *);
+
+static void
+ctk_banner_get_preferred_height(GtkWidget *, gint *, gint *);
+#else
 static gboolean
 ctk_banner_expose_event  (GtkWidget *, GdkEventExpose *);
 
 static void
 ctk_banner_size_request  (GtkWidget *, GtkRequisition *);
+#endif
 
 static gboolean
 ctk_banner_configure_event  (GtkWidget *, GdkEventConfigure *);
@@ -128,8 +140,14 @@ static void ctk_banner_class_init(
 
     gobject_class->finalize = ctk_banner_finalize;
 
+#ifdef CTK_GTK3
+    widget_class->draw = ctk_banner_draw_event;
+    widget_class->get_preferred_width  = ctk_banner_get_preferred_width;
+    widget_class->get_preferred_height = ctk_banner_get_preferred_height;
+#else
     widget_class->expose_event = ctk_banner_expose_event;
     widget_class->size_request = ctk_banner_size_request;
+#endif
     widget_class->configure_event = ctk_banner_configure_event;
 }
 
@@ -152,6 +170,25 @@ static void ctk_banner_finalize(
         g_object_unref(ctk_banner->background->pixbuf);
 }
 
+#ifdef CTK_GTK3
+static gboolean ctk_banner_draw_event(
+    GtkWidget *widget,
+    cairo_t *cr_i
+)
+{
+    CtkBanner *ctk_banner = CTK_BANNER(widget);
+    cairo_t *cr = gdk_cairo_create(gtk_widget_get_window(widget));
+
+    /* copy the backing pixbuf into the exposed portion of the window */
+
+    gdk_cairo_set_source_pixbuf(cr, ctk_banner->back.pixbuf, 0, 0);
+    cairo_paint(cr);
+
+    cairo_destroy(cr);
+
+    return FALSE;
+}
+#else
 static gboolean ctk_banner_expose_event(
     GtkWidget *widget,
     GdkEventExpose *event
@@ -180,6 +217,7 @@ static gboolean ctk_banner_expose_event(
 
     return FALSE;
 }
+#endif
 
 
 
@@ -221,9 +259,9 @@ static gboolean ctk_banner_configure_event(
 
     /* copy the base image into the backing pixbuf */
 
-    w = MIN(ctk_banner->background->w, ctk_banner->back.w);
-    h = MIN(ctk_banner->background->h, ctk_banner->back.h);
-    
+    w = NV_MIN(ctk_banner->background->w, ctk_banner->back.w);
+    h = NV_MIN(ctk_banner->background->h, ctk_banner->back.h);
+
 
     gdk_pixbuf_copy_area(ctk_banner->background->pixbuf,  // src
                          0,                               // src_x
@@ -321,15 +359,39 @@ static void ctk_banner_size_request(
 {
     CtkBanner *ctk_banner = CTK_BANNER(widget);
 
-    requisition->width = MAX(400,
-                             ctk_banner->logo->w +
-                             ctk_banner->artwork.w +
-                             ctk_banner->logo_pad_x +
-                             ctk_banner->artwork_pad_x);
-    
+    requisition->width = NV_MAX(400,
+                                ctk_banner->logo->w +
+                                ctk_banner->artwork.w +
+                                ctk_banner->logo_pad_x +
+                                ctk_banner->artwork_pad_x);
+
     requisition->height = ctk_banner->background->h;
 }
 
+#ifdef CTK_GTK3
+static void ctk_banner_get_preferred_width(
+    GtkWidget *widget,
+    gint *minimum_width,
+    gint *natural_width
+)
+{
+    GtkRequisition requisition;
+    ctk_banner_size_request(widget, &requisition);
+    *minimum_width = *natural_width = requisition.width;
+}
+
+static void ctk_banner_get_preferred_height(
+    GtkWidget *widget,
+    gint *minimum_height,
+    gint *natural_height
+)
+{
+    GtkRequisition requisition;
+    ctk_banner_size_request(widget, &requisition);
+    *minimum_height = *natural_height = requisition.height;
+}
+#endif
+
 
 
 /*
@@ -355,7 +417,6 @@ static gboolean select_artwork(BannerArtworkType artwork,
         { BANNER_ARTWORK_COLOR,          FALSE, 16, &color_pixdata          },
         { BANNER_ARTWORK_CONFIG,         FALSE, 16, &config_pixdata         },
         { BANNER_ARTWORK_CRT,            FALSE, 16, &crt_pixdata            },
-        { BANNER_ARTWORK_CURSOR_SHADOW,  FALSE, 16, &cursor_shadow_pixdata  },
         { BANNER_ARTWORK_DFP,            FALSE, 16, &dfp_pixdata            },
         { BANNER_ARTWORK_DISPLAY_CONFIG, FALSE, 16, &display_config_pixdata },
         { BANNER_ARTWORK_FRAMELOCK,      FALSE, 16, &framelock_pixdata      },
@@ -367,11 +428,12 @@ static gboolean select_artwork(BannerArtworkType artwork,
         { BANNER_ARTWORK_PENGUIN,        TRUE,  16, &penguin_pixdata        },
         { BANNER_ARTWORK_SDI,            FALSE, 16, &sdi_pixdata            },
         { BANNER_ARTWORK_SDI_SHARED_SYNC_BNC, FALSE, 16, &sdi_shared_sync_bnc_pixdata },
+        { BANNER_ARTWORK_SERVER_LICENSING, FALSE, 16, &server_licensing_pixdata },
         { BANNER_ARTWORK_SLIMM,          FALSE, 16, &slimm_pixdata          },
         { BANNER_ARTWORK_SOLARIS,        TRUE,  16, &solaris_pixdata        },
         { BANNER_ARTWORK_THERMAL,        FALSE, 16, &thermal_pixdata        },
-        { BANNER_ARTWORK_TV,             FALSE, 16, &tv_pixdata             },
         { BANNER_ARTWORK_VCS,            FALSE, 16, &vcs_pixdata            },
+        { BANNER_ARTWORK_VDPAU,          FALSE, 16, &vdpau_pixdata          },
         { BANNER_ARTWORK_X,              FALSE, 16, &x_pixdata              },
         { BANNER_ARTWORK_XVIDEO,         FALSE, 16, &xvideo_pixdata         },
         { BANNER_ARTWORK_SVP,            FALSE, 16, &svp_3dvp_pixdata       },
