@@ -54,7 +54,7 @@ G_BEGIN_DECLS
 
 
 /* Maximums */
-#define MAX_DEVICES 8  /* Max number of GPUs */
+#define MAX_DEVICES 64  /* Max number of GPUs */
 
 
 /* XF86VIDMODE */
@@ -143,6 +143,13 @@ typedef enum {
     MOSAIC_TYPE_BASE_MOSAIC_LIMITED,
 } MosaicType;
 
+typedef enum {
+    PIXELSHIFT_NONE = 0,
+    PIXELSHIFT_4K_TOP_LEFT,
+    PIXELSHIFT_4K_BOTTOM_RIGHT,
+    PIXELSHIFT_8K,
+} PixelShift;
+
 typedef struct nvSizeRec {
     int width;
     int height;
@@ -166,7 +173,7 @@ typedef struct nvModeLineRec {
 typedef struct nvSelectedModeRec {
     struct nvSelectedModeRec *next;
 
-    GtkWidget *label;       /* Label shown in dropdown menu */
+    gchar *text;            /* Text shown in dropdown menu */
 
     nvModeLinePtr modeline; /* Modeline this mode references */
 
@@ -215,6 +222,10 @@ typedef struct nvModeRec {
 
     Rotation rotation;
     Reflection reflection;
+    PixelShift pixelshift;
+
+    Bool forceCompositionPipeline;
+    Bool forceFullCompositionPipeline;
 
 } nvMode, *nvModePtr;
 
@@ -226,7 +237,7 @@ typedef struct nvDisplayRec {
     struct nvDisplayRec *next_in_screen;
     XConfigMonitorPtr    conf_monitor;
 
-    NvCtrlAttributeHandle *handle;      /* NV-CONTROL handle to device */
+    CtrlTarget         *ctrl_target;
 
     struct nvGpuRec    *gpu;            /* GPU the display belongs to */
     struct nvScreenRec *screen;         /* X screen the display is tied to */
@@ -271,7 +282,10 @@ typedef struct nvMetaModeRec {
     // Used for applying and generating metamodes (effective dimensions)
     GdkRectangle edim; /* Bounding box of all non-NULL modes */
 
-    char *string; /* Temp string used for modifying the metamode list */
+    /* Used to apply the metamode to the running X server */
+    char *cpl_str; /* metamode string from CPL */
+    char *x_str;   /* parsed CPL string from X */
+    char *x_str_entry; /* Points to string in metamode strings buffer */
 
 } nvMetaMode, *nvMetaModePtr;
 
@@ -292,7 +306,7 @@ typedef struct nvScreenRec {
      *
      */
 
-    NvCtrlAttributeHandle *handle;  /* NV-CONTROL handle to X screen */
+    CtrlTarget *ctrl_target;
     CtkEvent *ctk_event;
     int scrnum;
 
@@ -311,6 +325,8 @@ typedef struct nvScreenRec {
     int depth;      /* Depth of the screen */
     int stereo;     /* Stereo mode enabled on this screen */
     int overlay;    /* Overlay enabled on this screen */
+    int hw_overlay;
+    int ubb;
 
     nvDisplayPtr displays; /* List of displays using this screen */
     int num_displays; /* # of displays using this screen */
@@ -331,7 +347,6 @@ typedef struct nvScreenRec {
     Bool sli;
     char *sli_mode;
     char *multigpu_mode;
-    Bool dynamic_twinview;  /* This screen supports dynamic twinview */
     Bool no_scanout;        /* This screen has no display devices */
     Bool stereo_supported;  /* Can stereo be configured on this screen */
 
@@ -352,7 +367,7 @@ typedef struct GvoModeDataRec {
 typedef struct nvGpuRec {
     struct nvGpuRec *next_in_layout; /* List of all GPUs */
 
-    NvCtrlAttributeHandle *handle;  /* NV-CONTROL handle to GPU */
+    CtrlTarget *ctrl_target;
     CtkEvent *ctk_event;
 
     struct nvLayoutRec *layout; /* Layout this GPU belongs to */
@@ -390,7 +405,9 @@ typedef struct nvLayoutRec {
     XConfigLayoutPtr conf_layout;
     char *filename;
 
-    NvCtrlAttributeHandle *handle;
+    CtrlSystemList systems; /* Holds 1 system */
+    CtrlSystem *system;
+
 
     nvGpuPtr gpus;  /* Linked list of GPUs (next_in_layout) */
     int num_gpus;
@@ -465,18 +482,23 @@ typedef struct _CtkDisplayLayout
 {
     GtkVBox parent;
 
-    NvCtrlAttributeHandle *handle;
     CtkConfig             *ctk_config;
 
     GtkWidget   *drawing_area;   /* Drawing area */
     GtkWidget   *tooltip_area;   /* Tooltip area */
+#ifndef CTK_GTK3
     GtkTooltips *tooltip_group;  /* Tooltip group */
+#endif
 
     /* Layout configuration */
     nvLayoutPtr layout;
 
     /* Double buffering of layout image */
+#ifdef CTK_GTK3
+    cairo_t *c_context;
+#else
     GdkPixmap *pixmap;
+#endif
 
     /* Image information */
     GdkRectangle img_dim;
@@ -534,8 +556,7 @@ typedef struct _CtkDisplayLayoutClass
 
 GType       ctk_display_layout_get_type  (void) G_GNUC_CONST;
 
-GtkWidget*  ctk_display_layout_new       (NvCtrlAttributeHandle *,
-                                          CtkConfig *,
+GtkWidget*  ctk_display_layout_new       (CtkConfig *,
                                           nvLayoutPtr, /* Layout to display */
                                           int,         /* Width of image */
                                           int          /* Height of image */
