@@ -24,6 +24,7 @@
 #include "NVCtrl.h"
 #include <GL/glx.h> /* GLX #defines */
 #include <X11/extensions/Xrandr.h> /* Xrandr */
+#include <nvml.h>
 
 /* Make sure we are compiling with XRandR version 1.2 or greater */
 #define MIN_RANDR_MAJOR 1
@@ -77,6 +78,33 @@
 #define NV_DLSYM(handle, symbol) ({ dlerror(); dlsym(handle, symbol); })
 
 
+/* Useful macros to deal with target types */
+
+/* XXX Modify to be TRUE only for target types which are NVML-related */
+#define TARGET_TYPE_IS_NVML_COMPATIBLE(_TARGET_TYPE_) \
+    (((_TARGET_TYPE_) == GPU_TARGET) ||               \
+     ((_TARGET_TYPE_) == THERMAL_SENSOR_TARGET) ||    \
+     ((_TARGET_TYPE_) == COOLER_TARGET))
+
+#define TARGET_TYPE_NEEDS_NVCONTROL(_TARGET_TYPE_)     \
+    (!(TARGET_TYPE_IS_NVML_COMPATIBLE(_TARGET_TYPE_)))
+
+
+/* Useful macros to deal with attribute names */
+
+#define ATTRIBUTE_NAME(_ATTR_, _ATTR_TYPE_)                  \
+    ((nv_get_attribute_entry(_ATTR_, _ATTR_TYPE_) != NULL) ? \
+     (nv_get_attribute_entry(_ATTR_, _ATTR_TYPE_)->name) :   \
+     ("Unknown"))
+
+#define INT_ATTRIBUTE_NAME(_ATTR_) ATTRIBUTE_NAME(_ATTR_, CTRL_ATTRIBUTE_TYPE_INTEGER)
+#define STR_ATTRIBUTE_NAME(_ATTR_) ATTRIBUTE_NAME(_ATTR_, CTRL_ATTRIBUTE_TYPE_STRING)
+#define SOP_ATTRIBUTE_NAME(_ATTR_) ATTRIBUTE_NAME(_ATTR_, CTRL_ATTRIBUTE_TYPE_STRING_OPERATION)
+#define BIN_ATTRIBUTE_NAME(_ATTR_) ATTRIBUTE_NAME(_ATTR_, CTRL_ATTRIBUTE_TYPE_BINARY_DATA)
+#define COL_ATTRIBUTE_NAME(_ATTR_) ATTRIBUTE_NAME(_ATTR_, CTRL_ATTRIBUTE_TYPE_COLOR)
+#define CSC_ATTRIBUTE_NAME(_ATTR_) ATTRIBUTE_NAME(_ATTR_, CTRL_ATTRIBUTE_TYPE_SDI_CSC)
+
+
 typedef struct __NvCtrlAttributes NvCtrlAttributes;
 typedef struct __NvCtrlVidModeAttributes NvCtrlVidModeAttributes;
 typedef struct __NvCtrlAttributePrivateHandle NvCtrlAttributePrivateHandle;
@@ -87,6 +115,9 @@ typedef struct __NvCtrlXvTextureAttributes NvCtrlXvTextureAttributes;
 typedef struct __NvCtrlXvBlitterAttributes NvCtrlXvBlitterAttributes;
 typedef struct __NvCtrlXvAttribute NvCtrlXvAttribute;
 typedef struct __NvCtrlXrandrAttributes NvCtrlXrandrAttributes;
+typedef struct __NvCtrlNvmlAttributes NvCtrlNvmlAttributes;
+typedef struct __NvCtrlEventPrivateHandle NvCtrlEventPrivateHandle;
+typedef struct __NvCtrlEventPrivateHandleNode NvCtrlEventPrivateHandleNode;
 
 typedef struct {
     float brightness[3];
@@ -128,9 +159,38 @@ struct __NvCtrlXrandrAttributes {
     XRRCrtcGamma *pGammaRamp;
 };
 
+struct __NvCtrlNvmlAttributes {
+    struct {
+        void *handle;
+
+        typeof(nvmlInit)                           (*init);
+        typeof(nvmlShutdown)                       (*shutdown);
+        typeof(nvmlDeviceGetHandleByIndex)         (*deviceGetHandleByIndex);
+        typeof(nvmlDeviceGetUUID)                  (*deviceGetUUID);
+        typeof(nvmlDeviceGetCount)                 (*deviceGetCount);
+        typeof(nvmlDeviceGetTemperature)           (*deviceGetTemperature);
+        typeof(nvmlDeviceGetFanSpeed)              (*deviceGetFanSpeed);
+        typeof(nvmlDeviceGetName)                  (*deviceGetName);
+        typeof(nvmlDeviceGetVbiosVersion)          (*deviceGetVbiosVersion);
+        typeof(nvmlDeviceGetMemoryInfo)            (*deviceGetMemoryInfo);
+        typeof(nvmlDeviceGetPciInfo)               (*deviceGetPciInfo);
+        typeof(nvmlDeviceGetMaxPcieLinkGeneration) (*deviceGetMaxPcieLinkGeneration);
+        typeof(nvmlDeviceGetMaxPcieLinkWidth)      (*deviceGetMaxPcieLinkWidth);
+        typeof(nvmlDeviceGetVirtualizationMode)    (*deviceGetVirtualizationMode);
+
+    } lib;
+
+    unsigned int deviceIdx; /* XXX Needed while using NV-CONTROL as fallback */
+    unsigned int deviceCount;
+    unsigned int sensorCount;
+    unsigned int *sensorCountPerGPU;
+    unsigned int coolerCount;
+    unsigned int *coolerCountPerGPU;
+};
+
 struct __NvCtrlAttributePrivateHandle {
     Display *dpy;                   /* display connection */
-    int target_type;                /* Type of target this handle conrols */
+    CtrlTargetType target_type;     /* Type of target this handle controls */
     int target_id;                  /* screen num, gpu num (etc) of target */
 
     /* Common attributes */
@@ -141,14 +201,61 @@ struct __NvCtrlAttributePrivateHandle {
     NvCtrlXvAttributes *xv;         /* XVideo info */
     Bool glx;                       /* GLX extension available */
     NvCtrlXrandrAttributes *xrandr; /* XRandR extension info */
+
+    /* NVML-specific attributes */
+    NvCtrlNvmlAttributes *nvml;
 };
 
+struct __NvCtrlEventPrivateHandle {
+    Display *dpy;          /* display connection */
+    int fd;                /* file descriptor to poll for new events */
+    int nvctrl_event_base; /* NV-CONTROL base for indexing & identifying evts */
+    int xrandr_event_base; /* RandR base for indexing & identifying evts */
+};
+
+struct __NvCtrlEventPrivateHandleNode {
+    NvCtrlEventPrivateHandle *handle;
+    NvCtrlEventPrivateHandleNode *next;
+};
+
+
+/* Useful inline functions to deal with control targets */
+
+static inline Bool isTargetValid(const CtrlTarget *ctrl_target)
+{
+    return ((ctrl_target != NULL) && (ctrl_target->h != NULL));
+}
+
+static inline NvCtrlAttributePrivateHandle
+*getPrivateHandle(CtrlTarget *ctrl_target)
+{
+    if (!isTargetValid(ctrl_target)) {
+        return NULL;
+    }
+
+    return (NvCtrlAttributePrivateHandle *)(ctrl_target->h);
+}
+
+static inline const NvCtrlAttributePrivateHandle
+*getPrivateHandleConst(const CtrlTarget *ctrl_target)
+{
+    if (!isTargetValid(ctrl_target)) {
+        return NULL;
+    }
+
+    return (const NvCtrlAttributePrivateHandle *)(ctrl_target->h);
+}
+
+
 NvCtrlNvControlAttributes *
 NvCtrlInitNvControlAttributes (NvCtrlAttributePrivateHandle *);
 
 NvCtrlVidModeAttributes *
 NvCtrlInitVidModeAttributes (NvCtrlAttributePrivateHandle *);
 
+ReturnStatus
+NvCtrlFreeVidModeAttributes (NvCtrlAttributePrivateHandle *);
+
 
 /* Xv attribute functions */
 
@@ -158,9 +265,8 @@ NvCtrlInitXvAttributes (NvCtrlAttributePrivateHandle *);
 void
 NvCtrlXvAttributesClose (NvCtrlAttributePrivateHandle *);
 
-ReturnStatus
-NvCtrlXvGetStringAttribute (NvCtrlAttributePrivateHandle *,
-                           unsigned int, int, char **);
+ReturnStatus NvCtrlXvGetStringAttribute(const NvCtrlAttributePrivateHandle *,
+                                        unsigned int, int, char **);
 
 /* GLX extension attribute functions */
 
@@ -170,13 +276,11 @@ NvCtrlInitGlxAttributes (NvCtrlAttributePrivateHandle *);
 void
 NvCtrlGlxAttributesClose (NvCtrlAttributePrivateHandle *);
 
-ReturnStatus
-NvCtrlGlxGetVoidAttribute (NvCtrlAttributePrivateHandle *, unsigned int,
-                           int, void **);
+ReturnStatus NvCtrlGlxGetVoidAttribute(const NvCtrlAttributePrivateHandle *,
+                                       unsigned int, int, void **);
 
-ReturnStatus
-NvCtrlGlxGetStringAttribute (NvCtrlAttributePrivateHandle *, unsigned int,
-                             int, char **);
+ReturnStatus NvCtrlGlxGetStringAttribute(const NvCtrlAttributePrivateHandle *,
+                                         unsigned int, int, char **);
 
 
 /* XRandR extension attribute functions */
@@ -188,11 +292,11 @@ void
 NvCtrlXrandrAttributesClose (NvCtrlAttributePrivateHandle *);
 
 ReturnStatus
-NvCtrlXrandrGetStringAttribute (NvCtrlAttributePrivateHandle *,
-                                unsigned int, int, char **);
+NvCtrlXrandrGetStringAttribute(const NvCtrlAttributePrivateHandle *,
+                               unsigned int, int, char **);
 
 
-ReturnStatus NvCtrlXrandrGetColorAttributes(NvCtrlAttributePrivateHandle *h,
+ReturnStatus NvCtrlXrandrGetColorAttributes(const NvCtrlAttributePrivateHandle *h,
                                             float contrast[3],
                                             float brightness[3],
                                             float gamma[3]);
@@ -203,14 +307,16 @@ ReturnStatus NvCtrlXrandrSetColorAttributes(NvCtrlAttributePrivateHandle *h,
                                             float g[3],
                                             unsigned int bitmask);
 
-ReturnStatus NvCtrlXrandrGetColorRamp(NvCtrlAttributePrivateHandle *h,
+ReturnStatus NvCtrlXrandrGetColorRamp(const NvCtrlAttributePrivateHandle *h,
                                       unsigned int channel,
                                       uint16_t **lut,
                                       int *n);
 
+ReturnStatus NvCtrlXrandrReloadColorRamp(NvCtrlAttributePrivateHandle *h);
+
 /* XF86 Video Mode extension attribute functions */
 
-ReturnStatus NvCtrlVidModeGetColorAttributes(NvCtrlAttributePrivateHandle *h,
+ReturnStatus NvCtrlVidModeGetColorAttributes(const NvCtrlAttributePrivateHandle *h,
                                              float contrast[3],
                                              float brightness[3],
                                              float gamma[3]);
@@ -221,27 +327,34 @@ ReturnStatus NvCtrlVidModeSetColorAttributes(NvCtrlAttributePrivateHandle *h,
                                              float g[3],
                                              unsigned int bitmask);
 
-ReturnStatus NvCtrlVidModeGetColorRamp(NvCtrlAttributePrivateHandle *h,
+ReturnStatus NvCtrlVidModeGetColorRamp(const NvCtrlAttributePrivateHandle *h,
                                        unsigned int channel,
                                        uint16_t **lut,
                                        int *n);
 
-ReturnStatus
-NvCtrlVidModeGetStringAttribute (NvCtrlAttributePrivateHandle *,
-                                   unsigned int, int, char **);
+ReturnStatus NvCtrlVidModeReloadColorRamp(NvCtrlAttributePrivateHandle *h);
 
 ReturnStatus
-NvCtrlXrandrGetAttribute(NvCtrlAttributePrivateHandle *h,
+NvCtrlVidModeGetStringAttribute(const NvCtrlAttributePrivateHandle *,
+                                unsigned int, int, char **);
+
+ReturnStatus
+NvCtrlXrandrGetAttribute(const NvCtrlAttributePrivateHandle *h,
                          unsigned int display_mask, int attr, int64_t *val);
 
 /* Generic attribute functions */
 
-ReturnStatus
-NvCtrlNvControlQueryTargetCount(NvCtrlAttributePrivateHandle *, int, int *);
+NvCtrlAttributeHandle *NvCtrlAttributeInit(CtrlSystem *system,
+                                           CtrlTargetType target_type,
+                                           int target_id,
+                                           unsigned int subsystems);
 
 ReturnStatus
-NvCtrlNvControlGetAttribute (NvCtrlAttributePrivateHandle *, unsigned int,
-                             int, int64_t *);
+NvCtrlNvControlQueryTargetCount(const NvCtrlAttributePrivateHandle *, int,
+                                int *);
+
+ReturnStatus NvCtrlNvControlGetAttribute(const NvCtrlAttributePrivateHandle *,
+                                         unsigned int, int, int64_t *);
 
 ReturnStatus
 NvCtrlNvControlSetAttribute (NvCtrlAttributePrivateHandle *, unsigned int,
@@ -252,26 +365,31 @@ NvCtrlNvControlSetAttributeWithReply (NvCtrlAttributePrivateHandle *,
                                       unsigned int, int, int);
 
 ReturnStatus
-NvCtrlNvControlGetValidAttributeValues (NvCtrlAttributePrivateHandle *,
-                                        unsigned int, int,
-                                        NVCTRLAttributeValidValuesRec *);
+NvCtrlNvControlGetAttributePerms(const NvCtrlAttributePrivateHandle *,
+                                 CtrlAttributeType, int,
+                                 CtrlAttributePerms *);
+
+ReturnStatus
+NvCtrlNvControlGetValidAttributeValues(const NvCtrlAttributePrivateHandle *,
+                                       unsigned int, int,
+                                       CtrlAttributeValidValues *);
 
 ReturnStatus
 NvCtrlNvControlGetValidStringDisplayAttributeValues
-                                      (NvCtrlAttributePrivateHandle *,
+                                      (const NvCtrlAttributePrivateHandle *,
                                        unsigned int, int,
-                                       NVCTRLAttributeValidValuesRec *);
+                                       CtrlAttributeValidValues *);
 
 ReturnStatus
-NvCtrlNvControlGetStringAttribute (NvCtrlAttributePrivateHandle *,
-                                   unsigned int, int, char **);
+NvCtrlNvControlGetStringAttribute(const NvCtrlAttributePrivateHandle *,
+                                  unsigned int, int, char **);
 
 ReturnStatus
 NvCtrlNvControlSetStringAttribute (NvCtrlAttributePrivateHandle *,
-                                   unsigned int, int, const char *, int *);
+                                   unsigned int, int, const char *);
 
 ReturnStatus
-NvCtrlNvControlGetBinaryAttribute(NvCtrlAttributePrivateHandle *h,
+NvCtrlNvControlGetBinaryAttribute(const NvCtrlAttributePrivateHandle *h,
                                   unsigned int display_mask, int attr,
                                   unsigned char **data, int *len);
 
@@ -295,4 +413,31 @@ void NvCtrlAssignGammaInput(NvCtrlGammaInput *pGammaInput,
                             const float inGamma[3],
                             const unsigned int bitmask);
 
+/* NVML backend functions */
+
+NvCtrlNvmlAttributes *NvCtrlInitNvmlAttributes(NvCtrlAttributePrivateHandle *);
+void                  NvCtrlNvmlAttributesClose(NvCtrlAttributePrivateHandle *);
+
+ReturnStatus NvCtrlNvmlQueryTargetCount(const CtrlTarget *ctrl_target,
+                                        int target_type, int *val);
+ReturnStatus NvCtrlNvmlGetStringAttribute(const CtrlTarget *ctrl_target,
+                                          int attr, char **ptr);
+ReturnStatus NvCtrlNvmlSetStringAttribute(CtrlTarget *ctrl_target,
+                                          int attr, const char *ptr);
+ReturnStatus NvCtrlNvmlGetAttribute(const CtrlTarget *ctrl_target,
+                                    int attr, int64_t *val);
+ReturnStatus NvCtrlNvmlSetAttribute(CtrlTarget *ctrl_target, int attr,
+                                    int index, int val);
+ReturnStatus
+NvCtrlNvmlGetBinaryAttribute(const CtrlTarget *ctrl_target,
+                             int attr, unsigned char **data, int *len);
+ReturnStatus
+NvCtrlNvmlGetValidStringAttributeValues(const CtrlTarget *ctrl_target,
+                                        int attr,
+                                        CtrlAttributeValidValues *val);
+ReturnStatus
+NvCtrlNvmlGetValidAttributeValues(const CtrlTarget *ctrl_target,
+                                  int attr,
+                                  CtrlAttributeValidValues *val);
+
 #endif /* __NVCTRL_ATTRIBUTES_PRIVATE__ */
